
unit Envs;
{------This unit contains the hash table class used in the lisp interpreter
-------to store functions and variables }
interface

uses Declars,MiscFunc;

const HASHSIZE = 500;
type

{The envclass object is involved with manipulation of cells in the
 environments
 The base of the environments consists of a hash table of l. lists
 Each level is a node higher in the list from the base
 When a collision occurs the cells are externally chained to the base
 using the hashptr pointer}

    EnvClass = object
       h_table : array[0..HASHSIZE] of CellPtr;
       constructor Initialise;
       function HashFunction( Name : LispVar) : longint;
       function LookUp(Name : LispVar) : boolean;
       procedure Add(var Item : Cell;Name:LispVar);
       procedure Get(var Return : Cell; Name:LispVar);
       procedure DecLevel;
       private procedure NewVar(var Node :CellPtr;Item: Cell; Name :LispVar; Level:shortint);
    end;

var  environ : EnvClass;
implementation
uses Evaluate, Storage;
{--------------------------------=----------------------------------------}
constructor EnvClass.Initialise;
var i : integer;
begin
     for i := 0 to HASHSIZE do                  {Clean it up}
         h_table[i]:= nil;
end;
{--------------------------------=----------------------------------------}
function EnvClass.HashFunction( Name : LispVar) : longint;
var i:integer;
    Address:longint;
begin
     Address := 0;       {This function has a fair distribution
                          although it could be better}
     for i:= 1 to length(Name) do
        Address:= Address +  ord(Name[i]);

    HashFunction := Address mod HASHSIZE + 1;
end;
{--------------------------------=----------------------------------------}
procedure EnvClass.NewVar(var Node :CellPtr;Item: Cell; Name :LispVar; Level:shortint);
begin
  C_New(Node,Item);  {Create a new variable, link it, assign it level+ name}
  Node^.Name:=Name;
  Node^.Level:=Level;
  LinkToken(Node);
end;
{--------------------------------=----------------------------------------}
procedure EnvClass.Add(var Item : Cell ; Name:LispVar);
var Address , i  : integer;
    Node    : CellPtr;
begin
   Address := HashFunction(Name); {Get address }
   Node:= h_table[Address];       {Point to address}
   Item.Rubbish:=true;

   if h_table[Address] = nil then  {If new then create a new node}
   begin
          NewVar(h_table[Address],_NULL,Name,0);
          Node:= h_table[Address];
   end
   else if (Name <> h_table[Address]^.Name) then   {Check for collision}
      begin
         while ((Node^.Hashptr <> nil) and (Node^.Name <> Name)) do
         Node:=Node^.HashPtr;  {Traverse to appropriate link in chain}

        if ((Node^.Hashptr = nil) and (Node^.Name<> Name)) then    {If nil create a new link}
           begin
            NewVar(Node^.Hashptr,_NULL,Name,0);
            Node:=Node^.HashPtr;
         end;            {Continue normally as if nothing happened}
      end;

   for i := 1 to CurLevel do  {Traverse list to current level in table}
   begin
     if Node^.next=nil then  {If level doesn't exist create a new level}
          NewVar(Node^.Next,_NULL,Name,i);
     Node:= Node^.Next;
   end;

  Node^.Next:= nil;

 {If there is already a value stored , it is best to just dispose of it.
  This makes the algorithm MUCH more efficient as although it is
  longer but there are less holes in memory- this allows better memory
  efficiency}

 if Node^.Value <> nil then DisposeCons(Node^.value);

 C_New(Node^.Value,Item);   {Create the value item}
 LinkToken(Node^.Value);

 {If it is a global variable (ie at level 0) then protect it}
 if CurLevel= 0 then MarkCell(h_table[Address],false);
end;
{--------------------------------=----------------------------------------}
procedure EnvClass.Get(var Return : Cell; Name:LispVar);
var i      : integer;
    Node   : CellPtr;
    Address: longint;
begin
   Address := Hashfunction(Name);  {Get address}
   Node:= h_table[Address];        {Point to it}

    if Name <> h_table[Address]^.Name then   {In case of collision}
      begin
                                             {Traverse to appropriate node}
         while ((Node^.Hashptr<>nil) and (Node^.Name<> Name)) do
                Node:=Node^.HashPtr;
      end;

    if CurLevel > 0 then                {Move to appropriate level}
      for i:= 1 to CurLevel do
          Node:=Node^.Next;

    if Node<> nil then Return := Node^.Value^ {If something is present then}
         else Return := _NULL;                 {  return it}

end;
{--------------------------------=----------------------------------------}
function EnvClass.LookUp(Name : LispVar) : boolean;
var i       : integer;
    Node    : CellPtr;
    Address : longint;
begin
     Address := Hashfunction(Name); {Get address}
     Node := h_table[Address];      {Point to address}
  if h_table[Address]<> nil then
  begin
     if Name <> h_table[Address]^.Name then   {In case of collision}
      begin                                    {Traverse to approp. node}
         while ((Node<>nil) and (Node^.Name<> Name)) do
                Node:=Node^.HashPtr;
      end;

      if CurLevel > 0 then      {Move to approp. level}
         for i:= 1 to CurLevel do
               if Node <> nil then Node:=Node^.next;

                  {If something is present return true}
  end;
     LookUp := (Node <> nil);
end;
{--------------------------------=----------------------------------------}
procedure EnvClass.DecLevel;
var List    :        CellPtr;
begin
     dec(Level);  {Move down level}
     CurLevel:=Level;

     List:=TokenList;
     while List<>nil do   {for each argument...}
      begin
           if List^.level=Level then
            List^.Next:=nil;

           List:=List^.Protptr; {Next variable}
    end;
end;
{--------------------------------=----------------------------------------}
end.
