unit Evaluate;

interface

uses Declars,storage,Envs, MiscFunc,Errors;

{Procedure:Lookupsymbol
 Description: The Lookupsymbol function will search for a symbol in the current env.
              If none is found it will repeatedly move an environment
              down until the required symbol is found.  If no symbol is
              found a false is returned
 Arguments: The cell containing the name+ where result is placed
 Returns:true if found}
function LookUpSymbol(var Token: Cell):boolean;

{Procedure:Eval
 Description:The eval function is the most commonly used function in Lisp.  It
           evaluates a list of code and will direct program control to
           various function-  It returns a false if an error occured
 Arguments: The list of tokens to be evaluated
 Returns: true for no error, else false}
function Eval(var Token : cell): boolean;

implementation

uses Funcs,Uint;
{---------------------------------------------------------------------------}
function Eval(var Token : cell):boolean;
var i         :   integer;
    Args,
    Params,
    Funct     :   Cell;

begin
     if Token.Cell_Type = tNull then Exit;
     if (isAtom(Token)) then Exit;{Atoms return theirselves}

     if Token.Cell_Type = tSym then
     begin
       if LookupSymbol(Token)= false then   {Evaluate symbol}
        begin
             Error(4); Eval:=false;  exit;     {Unbound symbol}
        end;
     end
     else begin                                {Must be a function}
        Funct := Token.car^;
        if Tracing = true then
        begin
           printAtom(Funct);outstring(',');
        end;

        if Funct.Cell_Type= tSym then LookupSymbol(Funct);
        if (Funct.Cell_Type=tPredefined) then {Predefined Func execution}
         begin
               {Redirect to appropriate function}

              case (FindFuncIndex(Token.car^.Name)) of
                1:if SETQ(Token)= false then eval:=false;
                2:if LET(Token)= false then eval:=false;
                3:if DEFUN(Token)= false then eval:=false;
                4:if Quote(Token)= false then eval:=false;
                5:if Lambda(Token)= false then eval:=false;
                6:if  _if(Token)= false then eval:=false;
                7:if plus(token)= false then eval:=false;
                8:if minus(token)= false then eval:=false;
                9:if eq(token) = false then eval:=false;
               10:if greater(token) = false then eval:=false;
               11:if less(token) = false then eval:=false;
               12:if eqgreater(token) = false then eval:= false;
               13:if eqless(token) =false then eval:= false;
               14:if neq(token) = false then eval:= false;
               15:if times(token) = false then eval:= false;
               16:if division(token) = false then eval:= false;
               17:if cons(token) = false then Eval:= false;
               18:if Loop(Token) = false then Eval:= false;
               19:if _Until(Token) = false then Eval:=false;
               20:if _While(Token) = false then Eval:=false;
               21:if Car(Token) = false then Eval:= false;
               22:if Cdr(Token) = false then Eval:= false;
               23:if Atom(Token) = false then Eval:= false;
               24:if _AND(Token) = false then Eval:=false;
               25:if _OR(Token) = false then Eval:=false;
               26:if _NOT(Token) = false then Eval:=false;
               27:if Progn(Token) = false then Eval:=false;
               28:if Cond(Token) = false then Eval:=false;
               29:if _Sin(Token) = false then Eval:=false;
               30:if _Cos(Token) = false then Eval:=false;
               31:if _Tan(Token) = false then Eval:=false;
               32:if _ArcSin(Token) = false then Eval:=false;
               33:if _ArcCos(Token) = false then Eval:=false;
               34:if _ArcTan(Token) = false then Eval:=false;
               35:if Modulo(Token) = false then Eval:=false;
               36:if Pow(Token) = false then Eval:=false;
               37:if _Sqrt(Token) = false then Eval:=false;
               38:if Rand(Token) = false then Eval:=false;
               39:if Rnd(Token) = false then Eval:=false;
               40:if _Strlen(Token)= false then Eval:=false;
               41:if _Concat(Token)= false then Eval:=false;
               42:if Symbol(Token) = false then Eval:=false;
               43:if _String(Token) = false then Eval:=false;
               44:if SubString(Token) = false then Eval:=false;
               45:if nth(Token) = false then Eval:=false;
               46:if List(Token) = false then Eval:=false;
               47:if Null(Token) = false then Eval:=false;
               48:if Member(Token) = false then Eval:= false;
               49:if _Eval(Token) = false then Eval:= false;
               50:if Map(Token) = false then Eval:=false;
               51:if Apply(Token) = false then Eval:=false;
               52:if Setf(Token) = false then Eval:=false;
               53:if IsVar(Token)= false then Eval:=false;
               54:if IsStr(Token)= false then Eval:=false;
               55:if IsNum(Token)= false then Eval:=false;
               56:if IsSym(Token)= false then Eval:=false;
               57:if IsBool(Token)= false then Eval:=false;
               58:if Car(Token)= false then Eval:=false;
               59:if Cdr(Token)=false then Eval:=false;
               60:if IsCons(Token)=false then Eval:=false;
               61:if Princ(Token)=false then Eval:=false;
               62:if Trace(Token)=false then Eval:=false;
               63:if Search(Token)=false then Eval:=false;
           end;
     end
     else
       if (Funct.Cell_Type = tFunc) then {UserDefined Func execution}
       begin
                   {Local Variable Binding part}
          inc(Level);   {Move up a level on the environment}
          Curlevel:=Level;

          Args:= Funct.cdr^;  {Extract the arguments}
          Params:=Token.cdr^;

          if NumArgs(Args) <> NumArgs(Params) then
          begin
            Error(5); Eval:=false; exit;{No of parameters does not match no of args}
          end;

          for i:= 1 to NumArgs(Args) do
           begin
            Token:= Params.car^;      {Get each parameter}

           if Eval(Token) = false then         {Evaluate parameter}
              begin
                 Eval:=false; Exit;
              end;
           Environ.Add(Token , Args.car^.name);  {Add param into local env}
           Params := Params.cdr^;
           Args:=Args.cdr^;
        end;

 {Function Body execution}
    Funct:=Funct.car^;      {Get the function body}
    while Funct.Cell_Type <> tNull do
    begin
      Token:=Funct.car^;       {Get the current line}
      if (Eval(Token) = false) then      {Evaluate the line}
      begin
          Eval:=false;  CurLevel:= Level; exit;
      end;
      Funct:=Funct.cdr^;     {Move to next line}
   end;

    Environ.decLevel;   {Unchain down to a previous level}
   end else
    begin
       Error(4); eval:= false;   {Unbound Symbol}
    end;
  end;
  CurLevel:= Level;
end;
{---------------------------------------------------------------------------}
function LookUpSymbol(var Token: Cell):boolean;
var found    :            boolean;
begin
     CurLevel:=Level;         {Set the current to top of stack}
     found := false;
      while (found <> true) do
      begin
           if (Environ.LookUp(Token.Name)) then {if found at current level}
             begin
                  found:= true;
                  Environ.Get(Token,Token.Name);{Request the item}
             end
          else
               if CurLevel = 0 then break {Base level - still found nothing}
               else dec(CurLevel);    {Move down to the parent table}
      end;
      LookUpSymbol := found = true;
end;
{---------------------------------------------------------------------------}

end.