
unit Lexical;
{                          Lexical Analysis Stage               }
interface

uses Declars,Errors,MiscFunc,Storage;

var ch : char;  {The character currently being investigated}

{Procedure:ReadToken
 Description: Main entry point into lexical analyser
 Arguments: The variable into which the analysed tokens will be placed
 Returns: true for no errors, false for error}
function ReadToken(var Token : Cell):boolean;

{Procedure: Readlist
 Description: Reads and creates lisp conses
 Arguments:The variable into which the analysed tokens will be placed
 Returns: True for no errors, false for error}
function ReadList(var Token : Cell):boolean;

{Procedure: Finished
 Description: Predicate whether end of buffer has been reached
 Arguments: None
 Returns: True if end has been reached}
function Finished : boolean;

{Procedure: LookAhead
 Description: Reads the next character due to abiguities
 Arguments:NA
 Returns: The character}
function LookAhead:char;

{Procedure: ReadAtom
 Description: FSM which reads char by char converting them in tokens
 Arguments: The variable into which the token will be placed
 Returns: true for no error, false for error}
function ReadAtom( var Token: Cell):boolean;

{Procedure: ReadWord
 Description: Skip whitespaces
 Arguments:NA
 Returns: NA}
procedure ReadWord;

{Procedure: ReadChar
 Description: Read next char
 Arguments:NA
 Returns: NA}
procedure ReadChar;

implementation
{---------------------------------------------------------------------------}
function LookAhead:char;
begin
     LookAhead:=InputBuffer[InputPtr+1];
end;
{---------------------------------------------------------------------------}
function Finished : boolean;
begin
     Finished:=(InputPtr>=InputLen);
end;
{---------------------------------------------------------------------------}
procedure ReadChar;
begin
     inc(InputPtr);
     ch:=InputBuffer[InputPtr];
end;
{---------------------------------------------------------------------------}
procedure ReadWord;
begin
    ch := ' ';
    while ((ch = ' ') or (ch = #9)  or (ch = #0)  or
           (ch =#13)  or (ch =#10) and (Finished=false)) do
    begin
           Inc(InputPtr);
           ch := InputBuffer[InputPtr];
    end;
end;
{---------------------------------------------------------------------------}
function ReadToken(var Token : Cell):boolean;
var Temp               :      Cell;
begin
    case (ch) of    {Error - There is a superflous close parentheses}
     ')' : begin
                ReadWord;
                Error(2);ReadToken:=false;Exit; {Unexpected parentheses}
           end;
     '(' : begin
                ReadWord;

                if ReadList(Token)= false then   {Recursively read list}
                begin
                     ReadToken:=false;
                end;
           end;
     #39 :begin     {The ' symbol is a special case in lisp grammar}
              ReadChar;
              Temp:=_NULL;

              if ReadToken(Temp)= false then {lex. tokens following '}
              begin
                   ReadToken:= false;exit;
              end;

              CreateCons(Token,Temp,_NULL);
              CreateCons(Token,_QUOTE,Temp);
          end;
     ';' : begin  {Comment - Skip it}
            while ch <>#10 do  ReadChar;

            ReadWord;
            Token.Cell_Type:=tNull;
           end;
    else if ReadAtom(Token)= false then    {Its got to be a normal atom}
            begin
               ReadToken:=false; Exit;
            end;
   end;
end;
{---------------------------------------------------------------------------}
function ReadList(var Token : Cell):boolean;
var  car              :            Cell;
     cdr              :            Cell;
begin
   if Finished then   {Error - The lexer has finished reading but still}
     begin            {requires another bracket}
         Error(1);ReadList:=false ; Exit; {Expected brackets}
    end;

     if ch = ')' then  {Return a null ptr at the end of the list}
        Token:=_NULL
      else begin
         car := _NULL; {Initialise them- otherwise they fill up with garbage}
         cdr := _NULL;

         if (ReadToken(car)= false) then          {Read first atom (car)}
         begin
            ReadList:= false; Exit;
         end;

            {Expecting one of these}
         if ((ch<> ')') and (ch <> '(')) then ReadWord;

         if ReadList(cdr)= false then  {recursively Read Rest of list (cdr)}
         begin
              ReadList:= false; Exit;
         end;

         CreateCons(Token,car,cdr);
      end;
   ch:=' ';       {Reset character}
 end;
{---------------------------------------------------------------------------}
function ReadAtom(var Token: Cell):boolean;
var atom              :      string;
    Code              :      integer;
begin
{This is a  FSM to extract a string, integer,real or symbol from the input}
  atom:='';

  if (ch='"') then  { 1. Check for quotes}
   begin
     readchar;
     repeat
        atom:=atom+ ch;
        readchar;
        if Finished then
         begin
           Error(3); ReadAtom:= false; Exit; {Quotes not closed}
        end;
     until ch= '"';

     readchar;
     Token.value_str:=atom;
     Token.Cell_type := tString;
   end
  else if ((IsDigit(ch)) or     {2. Check for numbers pos. and neg}
       ((ch='-') and (isDigit(LookAhead)))) then
    begin
         if ch='-' then
         begin
            atom:=atom+ch;
            readchar;
         end;

       while (IsDigit(ch) and (Finished = false)) do
        begin
           atom:=atom+ch;
           readchar;
        end;
       val(atom,Token.value_num,Code);
       Token.Cell_Type := tNum;
    end

 else if (IsAlpha(ch) or isOp(ch)) then        {Got to be a symbol}
   begin
     while ((IsIdChar(ch)) and (Finished = False)) do
       begin
          atom:=atom + upcase(ch);
          readchar;
        end;
        Token.Name:= atom;
        Token.Cell_Type:= tSym;
     end;
  ReadAtom:=true;
 end;
end.
{---------------------------------------------------------------------------}
