unit UInt;

{$X+ $S-}
{$M 65520,0,655320}
{ The Quiklisp User Interface.  The unit is the main entry to the
  interpreter.  It is based on standard Borland Turbo Vision
 Units
    StdDlg    - Open file browser, change directory tree.
    MsgBox    - Simple dialog to display messages.
    HelpFile  - Context sensitive help.
    Editors   - Text Editor object.

  And of course this program uses many standard Turbo Vision
  objects and behaviors (menubar, desktop, status line, dialog boxes,
  mouse support, window resize/move/tile/cascade).}

interface

uses
  Dos, Objects,Drivers, Memory, Views, Menus, Dialogs, StdDlg, HistList,
  MsgBox, App, LispCmds, HelpFile, LispHelp , Editors,Strings,Declars,Storage,
  Lexical,Evaluate;

const
  HeapSize = 48 * (1024 div 16);  { Save 48k heap for main program }

var
  ClipWindow: PEditWindow;
  OpWindow :PEditWindow;
  Token:Cell;           {Lexically analysed linked list}
  ExeDir : DirStr;



{Procedure: PrintAtom
 Description: Outputs the appropriate value of the argument in the opwnd
 Arguments:  A single cell
 Returns:NA}
procedure PrintAtom(Token :cell);

{Procedure:OutString
 Description: Outputs a string to the opwnd
 Arguments: a string
 Returns:NA}
procedure OutString(Str : string);

{Procedure:OutNum
 Description:Outputs a number to the opwnd
 Arguments: the num to be output
 Returns: NA}
procedure OutNum(Val : real);

type
  { TLispInterface }
  PLispInterface = ^TLispInterface;
  TLispInterface = object(TApplication)
    constructor Init;
    procedure FileOpen(WildCard: PathStr);
    function OpenEditor(FileName: FNameStr; Visible: Boolean): PEditWindow;
    procedure GetEvent(var Event: TEvent); virtual;
    function  GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Idle; virtual;
    procedure InitMenuBar; virtual;
    procedure InitStatusLine; virtual;
    procedure OutOfMemory; virtual;
  end;


implementation
{Lisp related functions}
{---------------------------------------------------------------------------}
{Procedure: GetDir
 Description: Sets the exedir to the current directory
 Arguments:NA
 Returns: NA}

procedure GetDir;
var
  EXEName: PathStr;
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;
begin
  if Lo(DosVersion) >= 3 then EXEName := ParamStr(0)
  else EXEName := FSearch('Quiklisp.exe', GetEnv('PATH'));
  FSplit(EXEName, Dir, Name, Ext);
  ExeDir:=Dir;
end;

procedure OutString(Str : string);
var  EditWnd : PEditWindow;
     Out:array[0..64] of char;
begin
     StrPCopy(Out,Str);
     EditWnd :=PEditWindow(DeskTop^.Current);   {Get top window}
     OPWindow^.Editor^.InsertText(@Out,length(Str),false);
end;
{---------------------------------------------------------------------------}
procedure OutNum(Val : real);
var OutStr : LispString;
    EditWnd : PEditWindow;
    Out:array[0..64] of char;
begin
     EditWnd :=PEditWindow(DeskTop^.Current);   {Get top window}
     OutStr:='';
    if Val<=32767 then
       if int(Val)=Val then Str(trunc(Val),OutStr)
       else  Str(Val: 0 : 10, OutStr)
    else Str(Val: 0 : 10, OutStr);

    StrPCopy(Out,OutStr);
    OpWindow^.Editor^.InsertText(@Out,length(OutStr),false);
 end;
{---------------------------------------------------------------------------}
procedure PrintAtom(Token :cell);{Outputs the return value on the output wnd}
var Out: PChar;
begin
        case  Token.Cell_Type of
            tBool: begin
                        if Token.Value_Bool= true then OutString('T')
                        else OutString('NIL');
                   end;
            tNull   : OutString('NIL');
            tNum    : OutNum(Token.Value_Num);
            tString : OutString(Token.Value_str);
            tSym    : OutString(Token.Name);
            tFunc   : OutString(Token.Name);
            tPredefined: OutString(Token.Name);

            tCons : begin
                      OutString('(');
                      while Token.cell_Type <> tNull do
                      begin
                        PrintAtom(Token.car^); { For list outputs}
                        Token:=Token.cdr^;     {recursively call printatom}
                        if Token.Cell_Type<>tNull then OutString(' ');
                      end;
                      OutString(')');          {close the brackets}
                    end;
            end;
end;
{------------------------------------------------------------------------}
{Procedure:OpenOpWnd
 Description:Opens the opwndow which displays output
 Arguments: NA
 Returns:NA }

procedure OpenOpWnd;
var   P: PView; R: TRect;
      B: Word;
begin
if ((OpWindow= nil) or
 (OpWindow^.State<>sfActive)) then
 begin
   R.Assign(0,16,80,23);
     P := Application^.ValidView(New(PEditWindow,
    Init(R, EXEDIR+'LISP.OUT', wnNoNumber)));

  DeskTop^.Insert(P);
  OpWindow:= PEditWindow(P);
 end;
  OpWindow^.SetState(sfActive,true);
end;
{---------------------------------------------------------------------------}
{Procedure: ErrorProc
 Description:Takes care of errors occured during program, resets vars etc
 Arguments: NA
 Returns: NA}

procedure ErrorProc;
var ErrorFile : Text;
    i : integer;
    ErrorStr: string;
begin
    InputPtr := 0;InputLen := 0;  {Reset these}
    ErrorStr := ' ';
    Tracing:=false;

    Assign(ErrorFile,EXEDIR+'error.lst');  {Open error file}
    Reset(ErrorFile);
    for i:= 1 to ErrorMsg do        {Find message}
    begin
         Readln(ErrorFile,ErrorStr);
    end;

     Close(ErrorFile);            {Close file and..}
     OutString('Error ');
     OutNum(ErrorMsg);
     OutString(': ');
     OutString(ErrorStr);         {output string}
end;
{---------------------------------------------------------------------------}

{ CalcHelpName }

function CalcHelpName: PathStr;
var
  EXEName: PathStr;
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;
begin
  if Lo(DosVersion) >= 3 then EXEName := ParamStr(0)
  else EXEName := FSearch('Quiklisp.exe', GetEnv('PATH'));
  FSplit(EXEName, Dir, Name, Ext);
  if Dir[Length(Dir)] = '\' then Dec(Dir[0]);
  CalcHelpName := FSearch('LISP.HLP', Dir);

end;

function CreateFindDialog: PDialog;
var
  D: PDialog;
  Control: PView;
  R: TRect;
begin
  R.Assign(0, 0, 38, 12);
  D := New(PDialog, Init(R, 'Find'));
  with D^ do
  begin
    Options := Options or ofCentered;

    R.Assign(3, 3, 32, 4);
    Control := New(PInputLine, Init(R, 80));
    Insert(Control);
    R.Assign(2, 2, 15, 3);
    Insert(New(PLabel, Init(R, '~T~ext to find', Control)));
    R.Assign(32, 3, 35, 4);
    Insert(New(PHistory, Init(R, PInputLine(Control), 10)));

    R.Assign(3, 5, 35, 7);
    Insert(New(PCheckBoxes, Init(R,
      NewSItem('~C~ase sensitive',
      NewSItem('~W~hole words only', nil)))));

    R.Assign(14, 9, 24, 11);
    Insert(New(PButton, Init(R, 'O~K~', cmOk, bfDefault)));
    Inc(R.A.X, 12); Inc(R.B.X, 12);
    Insert(New(PButton, Init(R, 'Cancel', cmCancel, bfNormal)));

    SelectNext(False);
  end;
  CreateFindDialog := D;
end;

function CreateReplaceDialog: PDialog;
var
  D: PDialog;
  Control: PView;
  R: TRect;
begin
  R.Assign(0, 0, 40, 16);
  D := New(PDialog, Init(R, 'Replace'));
  with D^ do
  begin
    Options := Options or ofCentered;

    R.Assign(3, 3, 34, 4);
    Control := New(PInputLine, Init(R, 80));
    Insert(Control);
    R.Assign(2, 2, 15, 3);
    Insert(New(PLabel, Init(R, '~T~ext to find', Control)));
    R.Assign(34, 3, 37, 4);
    Insert(New(PHistory, Init(R, PInputLine(Control), 10)));

    R.Assign(3, 6, 34, 7);
    Control := New(PInputLine, Init(R, 80));
    Insert(Control);
    R.Assign(2, 5, 12, 6);
    Insert(New(PLabel, Init(R, '~N~ew text', Control)));
    R.Assign(34, 6, 37, 7);
    Insert(New(PHistory, Init(R, PInputLine(Control), 11)));

    R.Assign(3, 8, 37, 12);
    Insert(New(PCheckBoxes, Init(R,
      NewSItem('~C~ase sensitive',
      NewSItem('~W~hole words only',
      NewSItem('~P~rompt on replace',
      NewSItem('~R~eplace all', nil)))))));

    R.Assign(17, 13, 27, 15);
    Insert(New(PButton, Init(R, 'O~K~', cmOk, bfDefault)));
    R.Assign(28, 13, 38, 15);
    Insert(New(PButton, Init(R, 'Cancel', cmCancel, bfNormal)));

    SelectNext(False);
  end;
  CreateReplaceDialog := D;
end;

function DoEditDialog(Dialog: Integer; Info: Pointer): Word; far;
var
  R: TRect;
  T: TPoint;
begin
  case Dialog of
    edOutOfMemory:
      DoEditDialog := MessageBox('Not enough memory for this operation.',
        nil, mfError + mfOkButton);
    edReadError:
      DoEditDialog := MessageBox('Error reading file %s.',
        @Info, mfError + mfOkButton);
    edWriteError:
      DoEditDialog := MessageBox('Error writing file %s.',
        @Info, mfError + mfOkButton);
    edCreateError:
      DoEditDialog := MessageBox('Error creating file %s.',
        @Info, mfError + mfOkButton);
    edSaveModify:
      DoEditDialog := MessageBox('%s has been modified. Save?',
        @Info, mfInformation + mfYesNoCancel);
    edSaveUntitled:
      DoEditDialog := MessageBox('Save untitled file?',
        nil, mfInformation + mfYesNoCancel);
    edSaveAs:
      DoEditDialog := Application^.ExecuteDialog(New(PFileDialog, Init('*.lsp',
        'Save file as', '~N~ame', fdOkButton, 101)), Info);
    edFind:
      DoEditDialog := Application^.ExecuteDialog(CreateFindDialog, Info);
    edSearchFailed:
      DoEditDialog := MessageBox('Search string not found.',
        nil, mfError + mfOkButton);
    edReplace:
      DoEditDialog := Application^.ExecuteDialog(CreateReplaceDialog, Info);
    edReplacePrompt:
      begin
        { Avoid placing the dialog on the same line as the cursor }
        R.Assign(0, 1, 40, 8);
        R.Move((Desktop^.Size.X - R.B.X) div 2, 0);
        Desktop^.MakeGlobal(R.B, T);
        Inc(T.Y);
        if TPoint(Info).Y <= T.Y then
          R.Move(0, Desktop^.Size.Y - R.B.Y - 2);
        DoEditDialog := MessageBoxRect(R, 'Replace this occurence?',
          nil, mfYesNoCancel + mfInformation);
      end;
  end;
end;

{ TLispInterface }
constructor TLispInterface.Init;
var
  R: TRect;
  I: Integer;
  FileName: PathStr;
begin
  MaxHeapSize := HeapSize;
  inherited Init;
  RegisterObjects;
  RegisterViews;
  RegisterMenus;
  RegisterDialogs;
  RegisterApp;
  RegisterHelpFile;
  RegisterEditors;

  DisableCommands([cmSave, cmSaveAs, cmCut, cmCopy, cmPaste, cmClear,
    cmUndo, cmFind, cmReplace, cmSearchAgain, cmCloseAll]);

  EditorDialog := DoEditDialog;
  ClipWindow := OpenEditor('', False);
  if ClipWindow <> nil then
  begin
    Clipboard := ClipWindow^.Editor;
    Clipboard^.CanUndo := False;
  end;

  for I := 1 to ParamCount do
  begin
    FileName := ParamStr(I);
    if FileName[Length(FileName)] = '\' then
      FileName := FileName + '*.*';
    if (Pos('?', FileName) = 0) and (Pos('*', FileName) = 0) then
      OpenEditor(FExpand(FileName), True)
    else FileOpen(FileName);
  end;

     OpenOpWnd; {Display the output window}
end;

function TLispInterface.OpenEditor(FileName: FNameStr; Visible: Boolean): PEditWindow;
var
  P: PView;
  R: TRect;
  Ed: PEditWindow;
begin
  DeskTop^.GetExtent(R);
  R.B.Y:=R.B.Y-8;

  P := Application^.ValidView(New(PEditWindow,
    Init(R, FileName, wnNoNumber)));
  if not Visible then P^.Hide;
  DeskTop^.Insert(P);
  Ed := PEditWindow(P);
  OpenEditor:=Ed;
end;

procedure TLispInterface.FileOpen(WildCard: PathStr);
var
  FileName: FNameStr;
begin
  FileName := '*.lsp';
  if ExecuteDialog(New(PFileDialog, Init(WildCard, 'Open a file',
    '~N~ame', fdOpenButton + fdHelpButton, 100)), @FileName) <> cmCancel then
    OpenEditor(FileName, True);
end;

procedure TLispInterface.GetEvent(var Event: TEvent);
var
  W: PWindow;
  HFile: PHelpFile;
  HelpStrm: PDosStream;
const
  HelpInUse: Boolean = False;
begin
  inherited GetEvent(Event);
  case Event.What of
    evCommand:
      if (Event.Command = cmHelp) and not HelpInUse then
      begin
        HelpInUse := True;
        HelpStrm := New(PDosStream, Init(CalcHelpName, stOpenRead));
        HFile := New(PHelpFile, Init(HelpStrm));
        if HelpStrm^.Status <> stOk then
        begin
          MessageBox('Could not open help file.', nil, mfError + mfOkButton);
          Dispose(HFile, Done);
        end
        else
        begin
          W := New(PHelpWindow,Init(HFile, GetHelpCtx));
          if ValidView(W) <> nil then
          begin
            ExecView(W);
            Dispose(W, Done);
          end;
          ClearEvent(Event);
        end;
        HelpInUse := False;
      end;
    evMouseDown:
      if Event.Buttons <> 1 then Event.What := evNothing;
  end;
end;

function TLispInterface.GetPalette: PPalette;
const
  CNewColor = CAppColor + CHelpColor;
  CNewBlackWhite = CAppBlackWhite + CHelpBlackWhite;
  CNewMonochrome = CAppMonochrome + CHelpMonochrome;
  P: array[apColor..apMonochrome] of string[Length(CNewColor)] =
    (CNewColor, CNewBlackWhite, CNewMonochrome);
begin
  GetPalette := @P[AppPalette];
end;

procedure TLispInterface.HandleEvent(var Event: TEvent);

procedure ChangeDir;
var
  D: PChDirDialog;
begin
  D := New(PChDirDialog, Init(cdNormal + cdHelpButton, 101));
  D^.HelpCtx := hcFCChDirDBox;
  ExecuteDialog(D, nil);
end;
{Lisp related events}
{---------------------------------------------------------------------------}
{Procedure:RunProgram
 Description: Main entry level into interpreter, coordinates sub-programs
 Arguments:NA
 Returns: NA}

procedure RunProgram;

procedure GetText;
var EditWnd: PEditWindow;
begin
   EditWnd :=PEditWindow(DeskTop^.Current);   {Get top window}
   InputLen:=-1; InputPtr:=-1;

   if EditWnd<>OpWindow then         {The output wnd does not count!}
   begin
     if EditWnd^.Editor^.SelEnd > EditWnd^.Editor^.SelStart then
      begin       {If something is selected then}
        with  EditWnd^.Editor^ do
        begin
          InputBuffer:=@Buffer^[SelStart];  {Point to selected text}
          InputLen := SelEnd- SelStart;     {Get length of text}
        end;
     end;
   end;
end;

begin
  GetText;
  if InputLen <> -1 then           {Get text from buffer}
     begin
        ReadWord;          {Skip any whitespaces}
        repeat
             Token:=_NULL;
             if (ReadToken(Token) = false) then
             begin
                    ErrorProc;
                    OutString(#13#10);
             end
             else if Token.Cell_Type<> tNull then  {Skip evaluation of any comments}
             begin
                       if Eval(Token)=false then ErrorProc
                       else PrintAtom(Token);  {Output return token}
                       ReadWord;          {Skip the EOF chars}
                       OutString(#13#10);
             end;
             GarbageCollect;   {Clean the heap to be ready for next execution}
  until InputPtr>=InputLen;
  end;
 end;
{---------------------------------------------------------------------------}
procedure About;
var
  D: PDialog;
  Control: PView;
  R: TRect;
begin
  R.Assign(0, 0, 40, 11);
  D := New(PDialog, Init(R, 'About'));
  with D^ do
  begin
    Options := Options or ofCentered;

    R.Grow(-1, -1);
    Dec(R.B.Y, 3);
    Insert(New(PStaticText, Init(R,
      #13 +
      ^C'Quiklisp'#13 +
      #13 +
      ^C'Written by Andrew Duncan'#13 +
      #13 +
      ^C'For A Level project 2001')));

    R.Assign(15, 8, 25, 10);
    Insert(New(PButton, Init(R, 'O~K', cmOk, bfDefault)));
  end;
  if ValidView(D) <> nil then
  begin
    Desktop^.ExecView(D);
    Dispose(D, Done);
  end;
end;

procedure FileNew;
begin
  OpenEditor('', True);
end;

procedure ShowClip;
begin
  ClipWindow^.Select;
  ClipWindow^.Show;
end;

begin
  inherited HandleEvent(Event);
  case Event.What of
    evCommand:
      begin
        case Event.Command of
          cmOpen: FileOpen('*');
          cmNew: FileNew;
          cmShowClip: ShowClip;
          cmChangeDir: ChangeDir;
          cmAbout: About;
          cmRun  : RunProgram;    {Runs the program}
          cmOp   : OpenOpWnd;
        else
          Exit;
        end;
        ClearEvent(Event);
      end;
  end;
 end;

procedure TLispInterface.Idle;

function IsTileable(P: PView): Boolean; far;
begin
  IsTileable := (P^.Options and ofTileable <> 0) and
    (P^.State and sfVisible <> 0);
end;

begin
  inherited Idle;
  if Desktop^.FirstThat(@IsTileable) <> nil then
    EnableCommands([cmTile, cmCascade])
  else
    DisableCommands([cmTile, cmCascade]);
end;

procedure TLispInterface.InitMenuBar;
var
  R: TRect;
begin
  GetExtent(R);
  R.B.Y := R.A.Y+1;
  MenuBar := New(PMenuBar, Init(R, NewMenu(
    NewSubMenu('~F~ile', hcFile, NewMenu(
      StdFileMenuItems(nil)),
    NewSubMenu('~E~dit', hcEdit, NewMenu(
      StdEditMenuItems(
      NewLine(
      NewItem('~S~how clipboard', '', kbNoKey, cmShowClip, hcShowClip,
      nil)))),
    NewSubMenu('~S~earch', hcSearch, NewMenu(
      NewItem('~F~ind...', '', kbNoKey, cmFind, hcFind,
      NewItem('~R~eplace...', '', kbNoKey, cmReplace, hcReplace,
      NewItem('~S~earch again', '', kbNoKey, cmSearchAgain, hcSearchAgain,
      nil)))),
   NewSubMenu('Execu~t~e', hcNoContext, NewMenu(
      NewItem('~R~un','',kbF5,cmRun,hcNoContext,
      NewItem('~O~utput Window','',kbCtrlF5 ,cmOp,hcNoContext,nil))),

    NewSubMenu('~W~indow', hcWindows, NewMenu(
      StdWindowMenuItems(nil)),
    NewSubMenu('~H~elp',hcNoContext,NewMenu(
      NewItem('~O~nline Help guide','',kbF1,cmHelp,hcNoContext,
      NewItem('~A~bout Quiklisp','',kbNoKey,cmAbout,hcNoContext,
      nil))), nil)))))))));
end;

procedure TLispInterface.InitStatusLine;
var
  R: TRect;
begin
  GetExtent(R);
  R.A.Y := R.B.Y - 1;
  StatusLine := New(PStatusLine, Init(R,
    NewStatusDef(0, $FFFF,
      NewStatusKey('~Alt-X~ Exit', kbAltX, cmQuit,
      NewStatusKey('~F1~ Help', kbF1, cmHelp,
      NewStatusKey('~F3~ Open', kbF3, cmOpen,
      NewStatusKey('~Alt-F3~ Close', kbAltF3, cmClose,
      NewStatusKey('~Alt-F5~ Zoom', kbAltF5, cmZoom,
      NewStatusKey('~F5~ Execute',kbF5,cmRun,
      NewStatusKey('', kbF10, cmMenu,
      NewStatusKey('', kbCtrlF5, cmResize,
      nil)))))))),
    nil)));
end;

procedure TLispInterface.OutOfMemory;
begin
  MessageBox('Not enough memory available to complete operation.',
    nil, mfError + mfOkButton);
end;

begin
     GetDir;   {Find the execution directory and assign it to exedir}
end.
