unit Declars;
{     Main declarations and type definitions}
interface


const Max_Var_Length = 16;      {Length of variable}
      Max_String_Length= 32;    {Length of a string}

type
    LispString = string[Max_String_Length];
    LispVar    = string[Max_Var_Length];

    DataType = (tNull , tNum, tString,tBool,tCons, tSym, {Possible data types}
                tFunc, tPreDefined);

    CellPtr = ^CELL; {Pointer to cell}

 {The cell is the main data type used in Quiklisp.  It is used to move any
  data around the interpreter.  It is also used for storage on the heap}

   CELL = record
     Value        :      CellPtr; {Name of variable when bound}
     next         :      CellPtr;   {Environment shadowing cell}
     value_str    :      LispString; {String value}
     value_num    :      real;       {Number value}
     value_bool   :      boolean;   {Boolean value}
     car          :      CellPtr;   {Head of the list -Data part}
     cdr          :      CellPtr;   {Rest of the list -Pointer part}
     Cell_type    :      DataType;  {Indicates type of data being carried}
     Name         :      LispVar;   {Name of variable}
     Rubbish      :      boolean;   {Used for garbage collection}
     HashPtr      :      CellPtr;   {In case of collision}
     ProtPtr      :      CellPtr;   {Used during garbage collection}
     Address      :      CellPtr;   {The Source Address}
     Level        :      Byte;      {The level the var is chained at}
end;

const Num_Defined = 63;         {The predefined functions}
      PreDefined : array [1..Num_Defined] of LispVar=
                    ('SETQ','LET', 'DEFUN','QUOTE','LAMBDA',
                    'IF', '+','-','=','>','<','>=','<=','<>','*','/',
                    'CONS','LOOP','UNTIL','WHILE','CAR','CDR','ATOM',
                    'AND','OR','NOT','PROGN','COND',
                    'SIN', 'COS','TAN' ,'ASIN' , 'ACOS', 'ATAN','%',
                    'POWER', 'SQRT','RAND','RND',
                    'STRLEN', 'CONCAT' ,'SYMBOL', 'STRING' ,'SUBSTRING',
                    'NTH',  'LIST','NULL','MEMBER','EVAL', 'MAP','APPLY',
                    'SETF','ISVAR','ISSTR','ISNUM','ISSYM','ISBOOL',
                     'FIRST' , 'REST','ISCONS','PRINC','TRACE','SEARCH');

var _NULL       :       Cell;   {Constant cell declarations}
    _QUOTE      :       Cell;
    _TRUE       :       Cell;
    _FALSE      :       Cell;

    ErrorMsg    :       integer;     {Contains error code}
    Tracing     :       boolean;     {This is true when tracing is on}


{The level variable contains the level which the interpreter is working on}
{The curlevel variable is used since the interpreter will sometimes have to
 work at a lower environment}
    Curlevel,
    Level       :       integer;
    InputLen    :       integer;     {Length of input buffer}
    InputPtr    :       integer;    {Points to current character}
    InputBuffer :       PChar;  {Pointer to the selected text in buffer}

implementation
begin
     {Assignment of the above variables}
    with _NULL do
     begin
       Cell_type:= tNull;
       car := nil;  cdr := nil; next:= nil; Protptr:=nil;
       Rubbish:=true;
       Hashptr:=nil;
       Level:=0;
     end;

    _QUOTE:=_NULL;
    _QUOTE.Name:= 'QUOTE';
    _QUOTE.Cell_type:= tSym;

    _TRUE := _NULL;
    _TRUE.Cell_Type :=tBool;
    _TRUE.Value_bool:= true;

    _FALSE := _NULL;
    _FALSE.Cell_Type :=tBool;
    _FALSE.Value_bool:= false;

    tracing:=false;
end.
